/*
 * Nenad Taskovic 5/93
 * e-mail: neyt@fon.bg.ac.yu
 * www: http://fon.fon.bg.ac.yu/~neyt/
 */

#include <string.h>
#include <stdlib.h>
#include "liste.h"

#define OUT_OF_MEM 0

const char *errors[] = {
   "Out of memory"
};

extern void report_error (const char *);

List *init_list (unsigned int size)
{
   List *list;
   
   if ( (list = (List *) malloc(sizeof(List))) == NULL )
     report_error(errors[OUT_OF_MEM]);
   list->size = size;
   
   if ( (list->head = (Node *) malloc(sizeof(Node))) == NULL )
     report_error(errors[OUT_OF_MEM]);
   list->head->info = NULL;             /* pravi dummy node za head */
   list->head->prev = NULL;

   if ( (list->tail = (Node *) malloc(sizeof(Node))) == NULL )
     report_error(errors[OUT_OF_MEM]);
   list->tail->info = NULL;             /* pravi dummy node za tail */
   list->tail->next = NULL;
   
   list->head->next = list->tail;       /* povezuje dummy node-ove */
   list->tail->prev = list->head;
   
   return (list);
}

void destroy_list (List *list)
{
   Node *tmp;
   
   while (list->head) {
      tmp = list->head;
      list->head = list->head->next;
      free(tmp->info);
      free(tmp);
   }
   free(list);
}

Node *alloc_node (unsigned int size, void *info)
{
   Node *node;
   
   if ( (node = (Node *) malloc(sizeof(Node))) == NULL )
     report_error(errors[OUT_OF_MEM]);
   if ( (node->info = (void *) malloc(size)) == NULL )
     report_error(errors[OUT_OF_MEM]);
   memcpy(node->info, info, size);
   
   return (node);
}

void insert_before (Node *ref_node, Node *new_node) /* ref_node - reference node */
{
   new_node->next = ref_node;
   new_node->prev = ref_node->prev;
   ref_node->prev->next = new_node;
   ref_node->prev = new_node;
}

void insert_after (Node *ref_node, Node *new_node)
{
   new_node->prev = ref_node;
   new_node->next = ref_node->next;
   ref_node->next->prev = new_node;
   ref_node->next = new_node;
}

void insert_at_head (List *list, void *info)
{
   Node *new_node;
   
   new_node = alloc_node(list->size, info);
   insert_after(list->head, new_node);
}

void insert_at_tail (List *list, void *info)
{
   Node *new_node;
   
   new_node = alloc_node(list->size, info);
   insert_before(list->tail, new_node);
}

void insert_sorted (List *list, void *info, int (*compare)(const void *, const void *))
{
   Node *curr;
   Node *new_node;
   
   new_node = alloc_node(list->size, info);
   curr = list->tail->prev;
   while ( (curr != list->head) && (compare(curr->info, info) == 1) )
     curr = curr->prev;
   insert_after(curr, new_node);
}

int insert_unique_sorted (List *list, void *info, int (*compare)(const void *, const void *))
{
   Node *curr;
   Node *new_node;
   int cmp_res;
   
   curr = list->tail->prev;
   while ( (curr != list->head) && ((cmp_res = compare(curr->info, info)) == 1) )
     curr = curr->prev;
   if (cmp_res != 0) {
      new_node = alloc_node(list->size, info);
      insert_after(curr, new_node);
      return 1;
   }
   return 0;
}

void do_node (Node *node, int (*fdo)(void *))
{
   if (node)
     fdo(node->info);
}

void do_forward (List *list, int (*fdo)(void *))
{
   Node *curr;
   
   curr = list->head->next;
   while ( curr != list->tail && fdo(curr->info) )
      curr = curr->next;
}

void do_backward (List *list, int (*fdo)(void *))
{
   Node *curr;

   curr = list->tail->prev;
   while ( curr != list->head && fdo(curr->info) )
      curr = curr->prev;
}

void delete_node (Node *node)
{
   if (node) {
      node->prev->next = node->next;
      node->next->prev = node->prev;
      free(node->info);
      free(node);
   }
}

Node *find_node (List *list, void *key, int (*check)(void *, const void *))
{
   Node *curr;
   Node *found = NULL;
   int check_res;

   curr = list->head->next;
	while ( (curr != list->tail) && ((check_res = check(curr->info, key)) != RETURN) ) {
      if (check_res == FOUND)
	found = curr;
      curr = curr->next;
   }
   if (check_res == RETURN)
     found = curr;

   return (found);
}
